"""
.. module:: swag_client.schemas.validators
    :platform: Unix
.. author:: Kevin Glisson (kglisson@netflix.com)
.. author:: Mike Grima (mgrima@netflix.com)
"""
from marshmallow.validate import Validator
from marshmallow.exceptions import ValidationError


def validate_fqdn(n):
    if '.' not in n:
        raise ValidationError("{0} is not a FQDN".format(n))


class IsDigit(Validator):
    def __call__(self, value):
        if not value.isdigit():
            raise ValidationError("Must be a string of digits: {}".format(value))
        return True



import os
import sys
import json
import logging
from io import open

import jmespath

from swag_client.backend import SWAGManager

logger = logging.getLogger(__name__)

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


def load_file(data_file):
    """Tries to load JSON from data file."""
    try:
        with open(data_file, 'r', encoding='utf-8') as f:
            return json.loads(f.read())

    except JSONDecodeError as e:
        return []


def save_file(data_file, data):
    """Writes JSON data to data file."""
    with open(data_file, 'w', encoding='utf-8') as f:
        if sys.version_info > (3, 0):
            f.write(json.dumps(data))
        else:
            f.write(json.dumps(data).decode('utf-8'))


class FileSWAGManager(SWAGManager):
    def __init__(self, namespace, **kwargs):
        """Create a file based SWAG backend."""
        self.namespace = namespace
        self.version = kwargs['schema_version']

        if not kwargs.get('data_file'):
            self.data_file = os.path.join(kwargs['data_dir'], self.namespace + '.json')
        else:
            self.data_file = kwargs['data_file']

        if not os.path.isfile(self.data_file):
            logger.warning(
                'Backend file does not exist, creating... Path: {data_file}'.format(data_file=self.data_file)
            )

            save_file(self.data_file, [])

    def get(self, search_filter):
        """Fetch one item from file. Applying given filter, raises exception if more than one item found."""
        logger.debug('Fetching items. Filter: {filter}. Path: {data_file}'.format(
            filter=search_filter,
            data_file=self.data_file
        ))

        items = load_file(self.data_file)
        results = jmespath.search(search_filter, items)

        if len(results) > 1:
            raise Exception('Found more than one result for get!')

        elif len(results) == 1:
            return results[0]

    def create(self, item):
        """Creates a new item in file."""
        logger.debug('Creating new item. Item: {item} Path: {data_file}'.format(
            item=item,
            data_file=self.data_file
        ))

        items = self.get_all()
        if self.version == 1:
            items[self.namespace].append(item)
        else:
            items.append(item)

        save_file(self.data_file, items)
        return item

    def delete(self, item):
        """Deletes item in file."""
        logger.debug('Deleting item. Item: {item} Path: {data_file}'.format(
            item=item,
            data_file=self.data_file
        ))

        if self.version == 1:
            # TODO only supports aws providers
            items = self.get_all("{namespace}[?id!='aws-{id}']".format(
                id=item['id'],
                namespace=self.namespace
            ))
        else:
            items = self.get_all("[?id!='{id}']".format(id=item['id']))

        save_file(self.data_file, items)

    def update(self, item):
        """Updates item info in file."""
        logger.debug('Updating item. Item: {item} Path: {data_file}'.format(
            item=item,
            data_file=self.data_file
        ))
        self.delete(item)
        return self.create(item)

    def get_all(self, search_filter=None):
        """Gets all items in file."""
        logger.debug('Fetching items. Filter: {filter}. Path: {data_file}'.format(
            filter=search_filter,
            data_file=self.data_file
        ))

        items = load_file(self.data_file)

        # default values
        if not items:
            if self.version == 1:
                return {self.namespace: []}
            return []

        # filter values
        if search_filter:
            items = jmespath.search(search_filter, items)
            if self.version == 1:
                return {self.namespace: items}
            return items

        # return raw values
        return items



import sys
import json
import logging

import boto3
import jmespath
from botocore.exceptions import ClientError

from swag_client.backend import SWAGManager

logger = logging.getLogger(__name__)

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


def load_file(client, bucket, data_file):
    """Tries to load JSON data from S3."""
    logger.debug('Loading item from s3. Bucket: {bucket} Key: {key}'.format(
        bucket=bucket,
        key=data_file
    ))
    try:
        data = client.get_object(Bucket=bucket, Key=data_file)['Body'].read()

        if sys.version_info > (3,):
            data = data.decode('utf-8')

        return json.loads(data)
    except JSONDecodeError as e:
        return
    except ClientError as e:
        logger.exception(e)
        return


def save_file(client, bucket, data_file, items):
    """Tries to write JSON data to data file in S3."""
    logger.debug('Writing {number_items} items to s3. Bucket: {bucket} Key: {key}'.format(
        number_items=len(items),
        bucket=bucket,
        key=data_file
    ))
    return client.put_object(Bucket=bucket, Key=data_file, Body=json.dumps(items), ContentType='application/json', CacheControl='no-cache, no-store, must-revalidate')


class S3SWAGManager(SWAGManager):
    def __init__(self, namespace, **kwargs):
        """Create a S3 based SWAG backend."""
        self.namespace = namespace
        self.version = kwargs['schema_version']

        if kwargs.get('data_file'):
            self.data_file = kwargs['data_file']
        else:
            self.data_file = self.namespace + '.json'

        self.bucket_name = kwargs['bucket_name']
        self.client = boto3.client('s3', region_name=kwargs['region'])

    def get(self, search_filter):
        """Fetch one item from file. Applying given filter, raises exception if more than one item found."""
        logger.debug('Fetching items. Filter: {filter}. Path: {data_file}'.format(
            filter=search_filter,
            data_file=self.data_file
        ))

        items = load_file(self.client, self.bucket_name, self.data_file)
        results = jmespath.search(search_filter, items)

        if results:
            if len(results) > 1:
                raise Exception('Found more than one result for get!')

            elif len(results) == 1:
                return results[0]
        return results

    def create(self, item):
        """Creates a new item in file."""
        logger.debug('Creating new item. Item: {item} Path: {data_file}'.format(
            item=item,
            data_file=self.data_file
        ))

        items = self.get_all()
        if self.version == 1:
            items[self.namespace].append(item)
        else:
            items.append(item)

        save_file(self.client, self.bucket_name, self.data_file, items)
        return item

    def delete(self, item):
        """Deletes item in file."""
        logger.debug('Deleting item. Item: {item} Path: {data_file}'.format(
            item=item,
            data_file=self.data_file
        ))

        if self.version == 1:
            # TODO only supports aws providers
            items = self.get_all("{namespace}[?id!='{id}']".format(
                id=item['id'],
                namespace=self.namespace
            ))
        else:
            items = self.get_all("[?id!='{id}']".format(id=item['id']))

        save_file(self.client, self.bucket_name, self.data_file, items)

    def update(self, item):
        """Updates item info in file."""
        logger.debug('Updating item. Item: {item} Path: {data_file}'.format(
            item=item,
            data_file=self.data_file
        ))
        self.delete(item)
        return self.create(item)

    def get_all(self, search_filter=None):
        """Gets all items in file."""
        logger.debug('Fetching items. Filter: {filter}. Path: {data_file}'.format(
            filter=search_filter,
            data_file=self.data_file
        ))

        items = load_file(self.client, self.bucket_name, self.data_file)

        # default values
        if not items:
            if self.version == 1:
                return {self.namespace: []}
            return []

        # filter values
        if search_filter:
            items = jmespath.search(search_filter, items)
            if self.version == 1:
                return {self.namespace: items}
            return items

        # return raw values
        return items



from swag_client.migrations.versions import v2


def run_migration(data, version_start, version_end):
    """Runs migration against a data set."""
    items = []
    if version_start == 1 and version_end == 2:
        for item in data['accounts']:
            items.append(v2.upgrade(item))

    if version_start == 2 and version_end == 1:
        for item in data:
            items.append(v2.downgrade(item))
        items = {'accounts': items}
    return items



"""
SWAG Client
===========

Is a python client to interface with the SWAG service.
It also provides some helpful functions for dealing with accounts (aws or otherwise)
"""
import sys
import os.path

from setuptools import setup, find_packages

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

sys.path.insert(0, ROOT)

about = {}
with open(os.path.join(ROOT, "swag_client", "__about__.py")) as f:
    exec(f.read(), about)


install_requires = [
    'marshmallow>=2.13.5',
    'boto3>=1.4.6',
    'tabulate>=0.7.7',
    'beaker>=1.9.0',
    'click>=6.7',
    'jmespath>=0.9.3',
    'deepdiff>=3.3.0'
]

tests_require = [
    'pytest==3.1.3',
    'moto==1.0.1',
    'coveralls==1.1'
]

setup(
    name=about["__title__"],
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__uri__"],
    description=about["__summary__"],
    long_description='See README.md',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'tests': tests_require
    },
    entry_points={
        'console_scripts': [
            'swag-client = swag_client.cli:cli',
        ],
        'swag_client.backends': [
            'file = swag_client.backends.file:FileSWAGManager',
            's3 = swag_client.backends.s3:S3SWAGManager',
            'dynamodb = swag_client.backends.dynamodb:DynamoDBSWAGManager'
        ]
    },
    keywords=['aws', 'account_management']
)
